#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <math.h>
#include "cdtw.h"

/*	--------------------------------------------------------------------------------------------
	Creates a 2D array with contiguous memory blocks.

	@param src if not NULL, use the src memory block for the arr, else create a new memory block.
	@param rows rows of the array.
	@param cols cols of the array.
	@param sizeOfType size of an array element.

	@return Returns the 2D array.
*/
void** create2DArr(const void* src, const int rows, const int cols, const int sizeOfType)
{
	int i = rows;
	int step = sizeOfType*cols;
	void** arr = malloc(rows*sizeof(void*));
	unsigned char* it = (unsigned char*)(*arr = src>NULL?
												src:
												calloc(cols*rows,sizeOfType));
	unsigned char** it2 = (unsigned char**)(arr);

	for(;i--;it+=step)
		*it2++ = it;

	return arr;
}

/*	--------------------------------------------------------------------------------------------
	Free a 2D array with contiguous memory blocks and set array to NULL.

	@param arr array address.

	@return >0 if array memory not NULL and deallocated.
*/
int free2DArr(void*** arr)
{
	if(arr<=NULL) return 0;
	
	if(*arr<=NULL) 
	{
		*arr=NULL; 
		return -1; 
	}

	if(**arr<=NULL)
	{
		free(*arr);
		*arr=NULL;
		return -2;
	}

	free(**arr);
	free(*arr);
	*arr=NULL;
	return 1;
}

/*	--------------------------------------------------------------------------------------------
	Creates a cumulative cost matrix with associated shortest path in a top-down fashion.

	@param costMatrix input costMatrix.
	@param rows rows of the costMatrix.
	@param cols cols of the costMatrix.

	@return Returns the cumulative cost matrix.
*/
cdtwCC** createTopDownCumCostMatrix(const MAT_TYPE** costMatrix,  const int rows, const int cols)
{
	int i,j,k;
	int size = 2*cols*rows;
	cdtwCC** cumCostMatrix = (cdtwCC**)create2DArr(NULL,rows,2*cols, sizeof(cdtwCC));
	cdtwCC*p,*p1,*p2,*p3;
	
	/* init matrix with 0s */
	memset(*cumCostMatrix,0,size*sizeof(cdtwCC));

	/* init top row */
	for(j=cols-1,k=2*cols-1,i=rows-1;
		j>=0;
		j--,k--)
	{
		p = &(cumCostMatrix[i][k]);
		p->cumCost	 = costMatrix[i][j];
		p->cumRows = 1;
		p->cumCols	 = 1;
	}

	for(j=cols-1;
		j>=0;
		j--,k--)
	{
		p  = &(cumCostMatrix[i][k]);
		p1 = p+1;
		p->cumCost	 = costMatrix[i][j] + p1->cumCost;
		p->cumRows = 1;
		p->cumCols	 = p1->cumCols + 1;
		p->next = p1;
	}

	/* init right col */
	for(j=cols-1,k=2*cols-1,i=rows-2;
		i>=0;
		i--)
	{
		p  = &(cumCostMatrix[i][k]);
		p2 = p+cols;
		p->cumCost   = costMatrix[i][j] + p2->cumCost;
		p->cumCols  = 1;
		p->cumRows = p2->cumRows + 1;
		p->next = p2;
	}

	/* get min cum cost path using linear programming */
	for(i=rows-2;i>=0;i--)
	{
		for(k=2*cols-2;k>=0;k--)
		{
			j=k<cols?k:k-cols;

			p  = &(cumCostMatrix[i][k]);	/* current point */
			p1 = p+1;						/* right point */
			p2 = p+cols;					/* top point*/
			p3 = p2+1;						/* diag point */

			/* if right point has a lower cum cost */
			if( p1->cumCost < p2->cumCost && 
				p1->cumCost < p3->cumCost)
			{
				p->cumCost	 = costMatrix[i][j] + p1->cumCost;
				p->cumCols	 = p1->cumCols + 1;
				p->cumRows = p1->cumRows;
				p->next = p1;
			}
			/* if top point has a lower cum cost */
			else if(p2->cumCost < p1->cumCost && 
					p2->cumCost < p3->cumCost)
			{
				p->cumCost	 = costMatrix[i][j] + p2->cumCost;
				p->cumCols	 = p2->cumCols;
				p->cumRows = p2->cumRows + 1;
				p->next = p2;
			}
			/* if top right point has a lower cum cost */
			else
			{
				p->cumCost	 = costMatrix[i][j] + p3->cumCost;
				p->cumCols	 = p3->cumCols + 1;
				p->cumRows = p3->cumRows + 1;
				p->next = p3;
			}
		}
	}

	return cumCostMatrix;
}

/*	--------------------------------------------------------------------------------------------
	Creates a cumulative cost matrix with associated shortest path in a bottom-up fashion.

	@param costMatrix input costMatrix.
	@param cols cols of the costMatrix.
	@param rows rows of the costMatrix.

	@return Returns the cumulative cost matrix.
*/
cdtwCC** createBottomUpCumCostMatrix(const MAT_TYPE** costMatrix,  const int rows, const int cols)
{
	int i,j,k;
	int size = 2*cols*rows;
	cdtwCC** cumCostMatrix = (cdtwCC**)create2DArr(NULL,rows,2*cols, sizeof(cdtwCC));
	cdtwCC*p,*p1,*p2,*p3;
	
	/* init matrix with 0s */
	memset(*cumCostMatrix,0,size*sizeof(cdtwCC));

	/* init bottom row */
	for(j=0,k=0,i=0;
		j<cols;
		j++,k++)
	{
		p = &(cumCostMatrix[i][k]);
		p->cumCost	 = costMatrix[i][j];
		p->cumRows = 1;
		p->cumCols	 = 1;
	}

	for(j=0;
		j<cols;
		j++,k++)
	{
		p  = &(cumCostMatrix[i][k]);
		p1 = p-1;
		p->cumCost	 = costMatrix[i][j] + p1->cumCost;
		p->cumRows = 1;
		p->cumCols	 = p1->cumCols + 1;
		p->next = p1;
	}

	/* init left col */
	for(j=0,k=0,i=1;
		i<rows;
		i++)
	{
		p  = &(cumCostMatrix[i][k]);
		p2 = p-cols;
		p->cumCost = costMatrix[i][j] + p2->cumCost;
		p->cumCols  = 1;
		p->cumRows = p2->cumRows + 1;
		p->next = p2;
	}

	/* get min cum cost path using linear programming */
	for(i=1;i<rows;i++)
	{
		for(k=1;k<2*cols-1;k++)
		{
			j=k<cols?k:k-cols;
			p  = &(cumCostMatrix[i][k]);	/* current point */
			p1 = p-1;						/* left point */
			p2 = p-cols;					/* bottom point*/
			p3 = p2-1;						/* diag point */

			/* if left point has a lower cum cost */
			if( p1->cumCost < p2->cumCost && 
				p1->cumCost < p3->cumCost)
			{
				p->cumCost	 = costMatrix[i][j] + p1->cumCost;
				p->cumCols	 = p1->cumCols + 1;
				p->cumRows = p1->cumRows;
				p->next = p1;
			}
			/* if bottom point has a lower cum cost */
			else if(p2->cumCost < p1->cumCost && 
					p2->cumCost < p3->cumCost)
			{
				p->cumCost	 = costMatrix[i][j] + p2->cumCost;
				p->cumCols	 = p2->cumCols;
				p->cumRows = p2->cumRows + 1;
				p->next = p2;
			}
			/* if bottom left point has a lower cum cost */
			else
			{
				p->cumCost	 = costMatrix[i][j] + p3->cumCost;
				p->cumCols	 = p3->cumCols + 1;
				p->cumRows = p3->cumRows + 1;
				p->next = p3;
			}
		}
	}

	return cumCostMatrix;
}

/*	--------------------------------------------------------------------------------------------
	Find the starting point for CDTW and the max cost.

	@param costMatrix input costMatrix.
	@param rows rows of the costMatrix.
	@param cols cols of the costMatrix.
	@param radius defines the neighborhood in the search for the best starting point
	@param starti address of output starting row
	@param startj address of output starting column

	@return Returns max cost.
*/
MAT_TYPE findStartingPoint(const MAT_TYPE** costMatrix,  const int rows, const int cols,
						   int radius,
						   int* starti, int* startj)
{
	int minRadius = 5>radius?radius:5;
	int i,j,k,l,i2,j2;
	cdtwCC** topDownCumCostMatrix  = createTopDownCumCostMatrix(costMatrix, rows, cols);
	cdtwCC** bottomUpCumCostMatrix = createBottomUpCumCostMatrix(costMatrix, rows, cols);
	MAT_TYPE** totalCost = (MAT_TYPE**)create2DArr(NULL, rows,2*cols, sizeof(MAT_TYPE));
	MAT_TYPE* it = (MAT_TYPE*)(*costMatrix);
	cdtwCC* next;
	cdtwCC* p1,*p2;
	MAT_TYPE* c;
	int cumCols,cumRows;

	MAT_TYPE maxCost = (MAT_TYPE)-1e100;
	MAT_TYPE startCumCost = (MAT_TYPE)1e100;
	MAT_TYPE cumCost;

	/* find maxCost */
	for(i=cols*rows;i--;it++)
		if(*it>maxCost)
			maxCost=*it;

	/* find total cost */
#ifdef PRINT_DEBUG
	printf("     --> find total cost\n");
#endif
	
	for(i=0;i<rows;i++)
	{
		for(j=0;j<2*cols;j++)
		{
			/* total cost is the sum of the 2 semi-path costs */
			c = &(totalCost[i][j]);
			p1 = &(topDownCumCostMatrix[i][j]);
			p2 = &(bottomUpCumCostMatrix[i][j]);
			*c = p1->cumCost + p2->cumCost;

			/* if the path end too soon, pad with max cost */
			cumCols = p1->cumCols + p2->cumCols;
			cumRows= p1->cumRows + p2->cumRows;
			if(cumCols<cols) *c += (cols-cumCols)*maxCost;
			if(cumRows<rows) *c += (rows-cumRows)*maxCost;
		}
	}

	/* find starting point */
#ifdef PRINT_DEBUG
	printf("     --> find point\n");
#endif

	*starti = *startj = 0;
	for(i=0;i<rows;i++)
	{
		for(j=0;j<2*cols;j++)
		{
			/* compute the avg final cost along each path in a neighborhood of length=radius */
			l=1;
			cumCost = totalCost[i][j];
			
			/* path from top down cum matrix */
			next = topDownCumCostMatrix[i][j].next;
			for(k=radius;k-- && next>NULL;next = next->next)
			{
				/* find pos of the next point in the path */
				i2 = (int)(next - *topDownCumCostMatrix)/(2*cols);
				j2 = (int)(next - *topDownCumCostMatrix) - i2*2*cols;
				
				/* add the cost */
				cumCost += totalCost[i2][j2];
				l++;
			}

			/* path from bottom up cum matrix */
			next = bottomUpCumCostMatrix[i][j].next;
			for(k=radius;k-- && next>NULL;next = next->next)
			{
				/* find pos of the next point in the path */
				i2 = (int)(next - *bottomUpCumCostMatrix)/(2*cols);
				j2 = (int)(next - *bottomUpCumCostMatrix) - i2*2*cols;

				/* add the cost */
				cumCost += totalCost[i2][j2];
				l++;
			}

			cumCost/=l;

			/* record the smallest cost and its position */
			if(l>=minRadius && startCumCost>=cumCost)
			{
				*starti = i;
				*startj = j;
				startCumCost = cumCost;
			}
		}
	}

	if(*starti >= rows) (*starti) -= rows;
	if(*startj >= cols) (*startj) -= cols;

	/* free memory */
	free2DArr(&topDownCumCostMatrix);
	free2DArr(&bottomUpCumCostMatrix);
	free2DArr(&totalCost);

	return maxCost;
}
/*	--------------------------------------------------------------------------------------------
	Creates a cumulative cost matrix with associated shortest path in a bottom-up fashion.
	The matrix is not cyclic and the starting point is specified as a input.

	@param costMatrix input costMatrix.
	@param rows rows of the costMatrix.
	@param cols cols of the costMatrix.
	@param starti starting row
	@param startj starting column

	@return Returns the cumulative cost matrix.
*/
cdtwCC** createDTWCumCostMatrix(const MAT_TYPE** costMatrix,  const int rows, const int cols, 
								const int starti, const int startj)
{
	int i,j,k,l;
	int size = cols*rows;
	cdtwCC** cumCostMatrix = (cdtwCC**)create2DArr(NULL,rows,cols, sizeof(cdtwCC));
	cdtwCC*p,*p1,*p2,*p3;
	
	/* init matrix with 0s */
	memset(*cumCostMatrix,0,size*sizeof(cdtwCC));

	/* init end point */
	i=rows-1 + starti;	if(i>=rows)i-=rows;
	j=cols -1 + startj;	if(j>=cols) j-=cols;
	p = &(cumCostMatrix[i][j]);
	p->cumCost	 = costMatrix[i][j];
	p->cumCols	 = 1;
	p->cumRows = 1;
	p->next = NULL;

	/* init top row */
	for(l=cols-2;
		l>=0;
		l--)
	{
		j = l+startj;	if(j>=cols)j-=cols;
		p = &(cumCostMatrix[i][j]);
		p1 = j<cols-1?p+1:p+1-cols;
		p->cumCost	 = costMatrix[i][j] + p1->cumCost;
		p->cumCols	 = p1->cumCols + 1;
		p->cumRows = 1;
		p->next = p1;
	}

	/* init right col */
	j=cols-1 + startj;	if(j>=cols)j-=cols;
	for(k=rows-2;
		k>=0;
		k--)
	{
		i = k+starti;	if(i>=rows)i-=rows;
		p  = &(cumCostMatrix[i][j]);
		p2 = i<rows-1?p+cols:p+cols-size;
		p->cumCost   = costMatrix[i][j] + p2->cumCost;
		p->cumCols  = 1;
		p->cumRows = p2->cumRows + 1;
		p->next = p2;
	}

	/* get min cum cost path using linear programming */
	for(k=rows-2;k>=0;k--)
	{
		i = k+starti;	if(i>=rows)i-=rows;
		for(l=cols-2;l>=0;l--)
		{
			j = l+startj;	if(j>=cols)j-=cols;

			p  = &(cumCostMatrix[i][j]);			/* current point */
			p1 = j<cols-1?p+1:p+1-cols;			/* right point */
			p2 = i<rows-1?p+cols:p+cols-size;	/* top point*/
			p3 = j<cols-1?p2+1:p2+1-cols;			/* diag point */

			/* if right point has a lower cum cost */
			if( p1->cumCost < p2->cumCost && 
				p1->cumCost < p3->cumCost)
			{
				p->cumCost	 = costMatrix[i][j] + p1->cumCost;
				p->cumCols	 = p1->cumCols + 1;
				p->cumRows = p1->cumRows;
				p->next = p1;
			}
			/* if top point has a lower cum cost */
			else if(p2->cumCost < p1->cumCost && 
					p2->cumCost < p3->cumCost)
			{
				p->cumCost	 = costMatrix[i][j] + p2->cumCost;
				p->cumCols	 = p2->cumCols;
				p->cumRows = p2->cumRows + 1;
				p->next = p2;
			}
			/* if top right point has a lower cum cost */
			else
			{
				p->cumCost	 = costMatrix[i][j] + p3->cumCost;
				p->cumCols	 = p3->cumCols + 1;
				p->cumRows = p3->cumRows + 1;
				p->next = p3;
			}
		}
	}

	return cumCostMatrix;
}


/*	--------------------------------------------------------------------------------------------
	Find sequence alignment using CDTW.

	@param costMatrix input costMatrix.
	@param rows rows of the costMatrix (i.e. size of seq1).
	@param cols cols of the costMatrix (i.e. size of seq2).
	@param radius defines the neighborhood in the search for the best starting point
	@param idSeq1 address of the id array for sequence 1
	@param idSeq2 address of the id array for sequence 2
	@param nPairs address of the # of pairs (i.e. size of the id array)

	@return Returns the total cumulative cost.
*/
MAT_TYPE cdtw(const MAT_TYPE* costMatrix,  const int rows, const int cols, const int radius,
			  int** idSeq1, int** idSeq2, int* nPairs)
{
	int i,j;
	MAT_TYPE** costMatrix2D = (MAT_TYPE**)create2DArr(costMatrix,rows,cols, sizeof(MAT_TYPE));
	cdtwCC** cumCostMatrix;
	cdtwCC*p;
	int starti=0, startj=0;
	MAT_TYPE totalCost=0;
	int* id1,*id2;
	int maxNPairs = cols+rows+1;
	
	/* find starting point */
#ifdef PRINT_DEBUG
	printf("\n --> find start point\n");
#endif
	
	findStartingPoint(costMatrix2D, rows, cols,radius,&starti, &startj);

	/* create dtw cum matrix + path from starting point */
#ifdef PRINT_DEBUG
	printf(" --> dtw\n");
#endif
	
	cumCostMatrix = createDTWCumCostMatrix(costMatrix2D, rows, cols, starti, startj);
	p = &(cumCostMatrix[starti][startj]);
	totalCost = p->cumCost;

	/* allocate memory */
	*nPairs = 0;
	*idSeq1 = id1 = (int*)malloc(maxNPairs*sizeof(int));
	*idSeq2 = id2 = (int*)malloc(maxNPairs*sizeof(int));

	/* back track to get sequence alignment */
#ifdef PRINT_DEBUG
	printf(" --> back track\n");
#endif
	
	for(;maxNPairs-- && (p=p->next)>NULL;)
	{
		i = (int)(p - *cumCostMatrix)/cols;
		j = (int)(p - *cumCostMatrix) - i*cols;
		*id1++ = i;
		*id2++ = j;
		(*nPairs) ++;
	}

	/* free memory */
	free(costMatrix2D);
	free2DArr(&cumCostMatrix);

	return totalCost;
}